/** 
 * PersonInheritanceDrinkBeer.java --- single inheritance in Java
 * At the same time explaining implementation pattern "roles as subclasses"
 * Roles are expressed as subclasses (Professor, Beginner, Student,
 * LonglifeStudent, Alumnus, ProfessorEmeritus) are subclasses of Person
 */

import java.util.*;

/**
 * @author Uwe Assmann
 * How different people share features (e.g., names), but drink and behave differently. 
 * Showing the semantics of the overriding of method drinkBeer().
 * @date 2013-04-20
 */
public class PersonInheritanceDrinkBeer {
    // Main program. 
    public static void main(String args[]) {
	// The variables pointing to the objects
	// (1) What does Prof. Aßmann do different than his students?
	Professor UweAssmann = new Professor("Uwe Aßmann");

	// uweAsPerson is a polymorphic variable ("shadow" of UweAssmann)
	// However, it needs sometimes casts to use it
	Person uweAsPerson = UweAssmann; // upcast to a superclass
	uweAsPerson.drinkBeer();

	// (2) Make Student Victoria Miller a student of Prof. Aßmann
	Beginner VictoriaMiller = new Beginner("Victoria Miller");
 	Person vicciAsPerson = VictoriaMiller; // upcast to a superclass
	vicciAsPerson.drinkBeer();

	// (3) register Vicci
	VictoriaMiller.startStudy(UweAssmann);
	VictoriaMiller.study();
	VictoriaMiller.drinkBeer();

	// (4) mature Vicci
	vicciAsPerson = UweAssmann.doExam(VictoriaMiller);
	vicciAsPerson.drinkBeer();

	// (5) exmatriculate Vicci
	vicciAsPerson = UweAssmann.unregisterStudent((Student)vicciAsPerson); // Alternative with downcast
	vicciAsPerson.drinkBeer();
	/***
	**/
    }
}


/** 
 * Base class Person.
 */
class Person {
    String name; // internal attribute name; not visible
    // Constuctor 
    public Person() { }
    public Person(String a_name) { setName(a_name); }
    // Attribute query: read the name of a Person
    String getName() { return name; }
    // This method uses the Class prototype to get the name of the class (class Class)
    String getClassName() { return getClass().getName(); }
    // This method returns name of the object with ":" and type
    String getTypedName() { return getName()+":"+getClass().getName(); }
    // Attribute modifier: change the name of a Person
    void setName(String a_name) { name = a_name; }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Limo.... Emptying Limo...");
    }
}

/**
 * Abstract intermediate class Employee to share some features
 */
abstract class Employee extends Person {
    int salary; 
    int getSalary() { return salary; }
    void setSalary(int a_salary) { salary = a_salary; }
    void raiseSalary() { salary = salary*103/100; }
    void changeSalary(int a_raise) { salary = salary+a_raise; }
}


class Professor extends Person {
    int numberOfStudents; 
    // Crowd of students OR alumni (polymorphic..)
    Person students[];

    // Constructors
    public Professor() { }
    public Professor(String a_name) { 
	setName(a_name);
	numberOfStudents = 0; 
	// Create array of 200 students
	students = new Student[200];
    }
    String getName() { return name; }
    int getNumberOfStudents() { return numberOfStudents; }
    void setNumberOfStudents(int a_numberOfStudents) { numberOfStudents = a_numberOfStudents; }
    void incrNumberOfStudents() { numberOfStudents++; }
    void registerStudent(Student aStudent) { 
	if (numberOfStudents < 200) {
	    System.out.println("Welcome, dear student "+aStudent.getName()+" to my course ST!...");
	    students[numberOfStudents] = aStudent; 
	    incrNumberOfStudents();
	    System.out.println(getName()+": I have now "+getNumberOfStudents()+" students in my course ST...");
	} else {
	    System.out.println("Sorry, course is full...");
	}
    }
    public Student doExam(Beginner aBeginner) { 
	Student newStudent;
	System.out.println(aBeginner.getTypedName()+" is being examined...      done...");
	for (int index = 0; index <200; index++) {
	    Student curStudent = (Student)students[index];
	    if ( curStudent.getName().equals(aBeginner.getName()) ) {
		// Student found; change from Beginner to Student
		System.out.println("   Student "+aBeginner.getTypedName()+" under Index "+index+" found");
		if (curStudent instanceof Beginner) {
		    students[index] = null;
		    newStudent = new Student(aBeginner.getName());
		    students[index] = newStudent;
		    System.out.println("---- Beginner "+aBeginner.getTypedName()+" matured");
		    return newStudent;
		} 
		// otherwise, no action required
	    }
	}
	// Student not found
	System.out.println("ERROR: Student "+aBeginner.getTypedName()+" not found in student list");
	return null;
    }
    public Alumnus unregisterStudent(Student aStudent) { 
	Alumnus alumnus;
	for (int index = 0; index <200; index++) {
	    Student curStudent = (Student)students[index];
	    // String comparison

	    if ( curStudent.getName().equals(aStudent.getName()) ) {
		// Student found; change to Alumnus
		System.out.println("   Student "+aStudent.getTypedName()+" under Index "+index+" found");
		students[index] = null;
		alumnus = new Alumnus(aStudent.getName());
		System.out.println("---- Student "+aStudent.getTypedName()+" exmatriculated");
		return alumnus;
	    }
	}
	// Student not found
	System.out.println("ERROR: Student "+aStudent.getTypedName()+" not found in student list");
	return null;
    }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Leffe.... Emptying Leffe...");
    }
} 

class Student extends Person {
    public Student() { }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Radeberger.... Emptying Radeberger...");
    }
    public Student(String a_name) { 
	name = a_name;
    }  
    public void startStudy(Professor prof) { 
	prof.registerStudent(this);
	System.out.println("---- Starting to study with Prof. "+prof.getName()+" in course ST...");
    }
    public void study() { 
	System.out.println("Working hard .... in course ST...");
	System.out.println("  Writing up questions...");
	System.out.println("  Asking questions...");
	System.out.println("Working hard .... in course ST...");
    }
}

class Beginner extends Student {
    public Beginner(String aName) { setName(aName); }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Clausthaler alkoholfrei.... Emptying Clausthaler alkoholfrei...");
    }
}

class LonglifeStudent extends Student {
    public LonglifeStudent(String aName) { setName(aName); }
    void drinkBeer() { 
	System.out.println(getTypedName()+" no more money for Radeberger....");
    }
}

class Alumnus extends Person {
    public Alumnus(String aName) { setName(aName); }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Erdinger.... Emptying Erdinger...");
    }
}

class ProfessorEmeritus extends Professor {
    public ProfessorEmeritus (String aName) { setName(aName); }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Radler alkoholfrei.... Emptying Radler alkoholfrei...");
    }
}
